# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT
"""
A Kervi controller is a class that acts upon input from users or events or the underlaying os.
"""

from kervi.spine import Spine
from kervi.utility.thread import KerviThread
from kervi.utility.component import KerviComponent
from kervi.hal import GPIO

class UISelectControllerInput(KerviComponent):
    r"""
    Select component for Kervi controller

    Usage:

    .. code-block:: python

        class CameraFrameRate(ControllerSelect):
            def __init__(self, controller):
            ControllerSelect.__init__(self, controller.controllerId+".framerate", "Framerate" , controller)
            self.addOption("5", "5 / sec")
            self.addOption("10", "10 / sec")
            self.addOption("15", "15 / sec", True)

        def change(self, selectedOptions):
            print("Frame rate changed", selectedOptions)

    """
    def __init__(self, select_id, name, controller):
        KerviComponent.__init__(self, select_id, "select", name)
        #self.spine = Spine()
        self.controller = controller

        self._persist_value = False
        self.options = []
        self.selected_options = []
        self.change_command = self.component_id + ".change"
        self.spine.register_command_handler(self.change_command, self._on_change_handler)
        self._ui_parameters = {
            "size": 0,
            "type": "dropdown",
            "link_to_header": False,
            "label_icon": None,
            "label" : self.name,
            "flat": False,
            "inline": False,
            "input_size": 50
        }
        self.controller.add_input(self)

    @property
    def input_id(self):
        return self.component_id
    
    @property
    def value(self):
        """Current value/state of the component"""
        return self.selected_options

    @value.setter
    def value(self, new_value):
        self._on_change_handler(new_value)

    @property
    def persist_value(self):
        """ If true the value is saved to kervi DB when it change"""
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        self._persist_value = do_persist

    def _load_persisted(self):
        if self.persist_value:
            self._on_change_handler(self.settings.retrieve_value("selected_options"), False)

    def link_to_dashboard(self, dashboard_id, panel_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.
             
            * *type* (``str``) -- Type of select box, dropdown or list.

            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.
                
            * *input_size* (``int``) -- width of the select box as a percentage of the total container it sits in.

        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            panel_id,
            **kwargs
            )

    def _get_info(self):
        return {
            "onSelect": self.change_command,
            "options": self.options,
        }

    def add_option(self, value, text, selected=False):
        """
        Add option to select
        """
        option = {"value": value, "text": text, "selected":selected}
        self.options += [option]
        if selected:
            self.selected_options += [option]
            if self._persist_value:
                self.settings.store_value("selected_options", self.selected_options)


    def _on_change_handler(self, selected_options, allow_persist=True):
        self.spine.log.debug(
            "controller select change:{0}/{1}",
            self.controller.component_id,
            self.component_id
        )

        for option in self.options:
            option["selected"] = False

        self.selected_options = []
        for selected_option in selected_options:
            for option in self.options:
                if option["value"] == selected_option:
                    option["selected"] = True
                    self.selected_options += [option]

        self.change(self.selected_options)

        if self._persist_value and allow_persist:
            self.settings.store_value("value", self.selected_options)

        self.spine.trigger_event(
            "controllerSelectChange",
            self.component_id,
            {"select":self.component_id, "value":self.options}
        )

    def change(self, selected_options):
        """ Abstract method fill in with your own code to be executed when a change occur. """
        self.controller.input_changed(self)

class UIButtonControllerInput(KerviComponent):
    def __init__(self, button_id, name, controller):
        KerviComponent.__init__(self, button_id, "button", name)
        self.controller = controller
       
        self.state = False
        self._down_command = self.component_id + ".down"
        self._up_command = self.component_id + ".up"
        self.spine.register_command_handler(self._down_command, self._on_down_handler)
        self.spine.register_command_handler(self._up_command, self._on_up_handler)
        self._ui_parameters = {
            "size": 0,
            "input_size": None,
            "type": "normal",
            "link_to_header": False,
            "label_icon": None,
            "label": None,
            "flat": False,
            "inline": False,
            "button_text": self.name,
            "button_icon": None

        }
        self.state = False
        self.controller.add_input(self)

    @property
    def input_id(self):
        return self._component_id

    @property
    def value(self):
        """Current state of the component"""
        return self.state

    @value.setter
    def value(self, new_value):
        if new_value:
            self._on_down_handler()
        else:
            self._on_up_handler()

    def _load_persisted(self):
        pass

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.
             
            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.
                
            * *input_size* (``int``) -- width of the select box as a percentage of the total container it sits in.
            
            * *button_text* (``str``) -- Text to display on button. Default is name of button.
            
            * *button_icon* (``str``) -- Icon to display on button.


        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            **kwargs
            )

    def _get_info(self):
        return {
            "onPress":self._down_command,
            "onRelease":self._up_command,
            "state":self.state
        }

    def _on_down_handler(self):
        self.spine.log.debug(
            "controller button pressed:{0}/{1}",
            self.controller.component_id,
            self.component_id
        )

        if not self.state:
            self.state = True
            self.down()

            self.spine.trigger_event(
                "controllerButtonStateChange",
                self.component_id,
                {"button":self.component_id, "state":self.state}
            )

    def _on_up_handler(self):
        self.spine.log.debug(
            "controller button released:{0}/{1}",
            self.controller.component_id,
            self.component_id
        )
        if self.state:
            self.state = False
            self.up()

            self.spine.trigger_event(
                "controllerButtonStateChange",
                self.component_id,
                {"button":self.component_id, "state":self.state}
            )

    def down(self):
        """
        Abstract method fill in with your own code that should be executed when button is pressed down.
        """
        self.controller.input_changed(self)

    def up(self):
        """
        Abstract method fill in with your own code that should be executed when button is released.
        """
        self.controller.input_changed(self)

class UISwitchButtonControllerInput(KerviComponent):
    """
    Switch button controller component, shows a on/off button in UI
    """
    def __init__(self, button_id, name, controller):
        KerviComponent.__init__(self, button_id, "switchButton", name)
        #self.spine = Spine()
        self.controller = controller

        self.state = False

        self.on_command = self.component_id + ".on"
        self.off_command = self.component_id + ".off"

        self.spine.register_command_handler(self.on_command, self._on_on_handler)
        self.spine.register_command_handler(self.off_command, self._on_off_handler)
        self._ui_parameters = {
            "size": 0,
            "input_size": None,
            "type": "normal",
            "link_to_header": False,
            "flat": False,
            "on_text": "On",
            "off_text": "Off",
            "label": self.name,
            "label_icon": None,
            "inline": False,
            "on_icon": None,
            "off_icon": None,
            "read_only": False
        }
        self._persist_value = False
        self.controller.add_input(self)

    @property
    def input_id(self):
        return self.component_id

    @property
    def value(self):
        """Current state of the component"""
        return self.state

    @value.setter
    def value(self, new_value):
        if new_value:
            self._on_on_handler()
        else:
            self._on_off_handler()

    @property
    def persist_value(self):
        """ If true the value is saved to kervi DB when it change"""
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        self._persist_value = do_persist

    def _load_persisted(self):
        if self.persist_value:
            state = self.settings.retrieve_value("state")
            if state:
                self._on_on_handler(allow_persist=False)
            else:
                self._on_off_handler(allow_persist=False)

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.
             
            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.
                
            * *input_size* (``int``) -- width of the switch as a percentage of the total container it sits in.
            
            * *on_text* (``bool``) -- Text to show on button when on.
            
            * *off_text* (``bool``) -- Text to show on button when off.
            
            * *on_icon* (``bool``) -- Icon to show on button when on.
            
            * *off_icon* (``bool``) -- Icon to show on button when off.
            
            * *read_only* (``bool``) -- If true the user is not able to change state.

        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            **kwargs
            )

    def _get_info(self):
        return {
            "onCommand":self.on_command,
            "offCommand":self.off_command,
            "state":self.state,
        }

    def _on_on_handler(self, allow_persist=True):
        self.spine.log.debug(
            "controller button on:{0}/{1}",
            self.controller.component_id,
            self.component_id
        )

        if not self.state:
            self.state = True
            self.on()

            if self._persist_value and allow_persist:
                self.settings.store_value("state", self.state)


            self.spine.trigger_event(
                "controllerButtonStateChange",
                self.component_id,
                {"button":self.component_id, "state":self.state}
            )

    def on(self):
        """
        Abstract method fill that is executed when button is turned on.
        """
        self.controller.input_changed(self)


    def _on_off_handler(self, allow_persist=True):
        self.spine.log.debug(
            "controller button off:{0}/{1}",
            self.controller.component_id,
            self.component_id
        )
        if self.state:
            self.state = False
            self.off()
            if self._persist_value and allow_persist:
                self.settings.store_value("value", self.state)

            self.spine.trigger_event(
                "controllerButtonStateChange",
                self.component_id,
                {"button":self.component_id, "state":self.state}
            )

    def off(self):
        """
        Abstract method that is executed then button is turned off.
        """
        self.controller.input_changed(self)

class DigitalGPIOControllerInput(UISwitchButtonControllerInput):
    """
    Digital GPIO input that listens to a digital input going high or low.
    This input will show up as an UISwitchButtonControllerInput when linked to a dashboard panel
    and all ui settings for UISwitchButtonControllerInput applies to this input.

    :param input_id:
            id of the input. This id used to reference this input in other parts of the kervi app.
    :type input_id: str

    :param name:
            Name of this component is used when displayed.
    :type name: str

    :param controller:
            The controller that this component belongs to.
    :type controller: Controller instance.

    :param channel:
            channel on the gpio device that should be used.
    :type channel: int.

    :param gpio_device:
            GPIO device to use. Defaults to the host platforms gpio e.g the Raspbery PI's gpio.
    
    :type gpio_device: IGPIODeviceDriver (defined in kervi.utility.hal.gpio).

    """
    def __init__(self, button_id, name, controller, channel, gpio_device=GPIO):
        UISwitchButtonControllerInput.__init__(self, button_id, name)
        self.channel = channel
        self.gpio = gpio_device
        self.set_ui_parameter("read_only", True)
        self.gpio.define_as_input(self.channel)
        self.gpio.listen(self.channel, self._on_edge)
        self.state = self.gpio.get(self.channel)


    @property
    def value(self):
        """Current state of the component"""
        return self.state

    @value.setter
    def value(self, new_value):
        pass

    def _load_persisted(self):
        pass

    def _on_edge(self, state):
        print("state", state)
        if self.gpio.get(self.channel):
            self.on_high()
            self.state = True
        else:
            self.on_low()
            self.state = False

        self.spine.trigger_event(
            "controllerButtonStateChange",
            self.component_id,
            {"button":self.component_id, "state":self.state}
        )

    def on_high(self):
        """
        Abstract method that is executed when the channel is going high.
        """
        self.controller.input_changed(self)

    def on_low(self):
        """
        Abstract method that is executed when the channel is going low.
        """
        self.controller.input_changed(self)


class UINumberControllerInput(KerviComponent):
    """
    A number input component shows as a slider on dashboards.
    """
    def __init__(self, input_id, name, controller):
        KerviComponent.__init__(self, input_id, "number-input", name)
        #self.spine = Spine()
        self.controller = controller
        self.min_value = -100
        self.max_value = 100
        self.unit = ""
        self._value = 0
        self.command = self.component_id + ".setValue"
        self.spine.register_command_handler(self.command, self._set_value)
        self._ui_parameters = {
            "size": 0,
            "input_size": 50,
            "type": "horizontal_slider",
            "link_to_header": False,
            "label_icon": None,
            "label": self.name,
            "flat": False,
            "inline": False,
            "read_only": False,
            "value_size":10
        }
        self._persist_value = False
        self.controller.add_input(self)

    @property
    def input_id(self):
        return self.component_id

    @property
    def value(self):
        """Current state of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        self._set_value(new_value)

    @property
    def persist_value(self):
        """ If true the value is saved to kervi DB when it change"""
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        self._persist_value = do_persist

    def _load_persisted(self):
        if self._persist_value:
            self._set_value(self.settings.retrieve_value("value"), False)

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.

            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.

            * *input_size* (``int``) -- width of the slider as a percentage of the total container it sits in.

            * *value_size* (``int``) -- width of the value area as a percentage of the total container it sits in.

            * *type* (``string``) -- 'vertical_slider' or 'horizontal_slider.
            
        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            **kwargs
            )

    def _set_value(self, nvalue, allow_persist=True):
        if self.value != nvalue:
            self.spine.log.debug(
                "value change on input:{0}/{1} value:{2}",
                self.controller.component_id,
                self.component_id,
                nvalue
            )
            old_value = self.value
            self._value = nvalue
            self.value_changed(nvalue, old_value)
            if self._persist_value and allow_persist:
                self.settings.store_value("value", self.value)

            self.spine.trigger_event(
                "changeControllerInputValue",
                self.component_id,
                {"input":self.component_id, "value":nvalue}
            )

    def value_changed(self, newValue, old_value):
        self.controller.input_changed(self)

    def _get_info(self):
        return {
            "unit":self.unit,
            "value":self.value,
            "maxValue":self.max_value,
            "minValue":self.min_value,
            "command":self.command,
        }

    def on_get_value(self):
        return self.value


class AnalogGPIOControllerInput(UINumberControllerInput):
    """
    Analog GPIO input that listens to the signal level on analog input.
    This input will show up as an UINumberControllerInput when linked to a dashboard panel
    and all ui settings for UINumberControllerInput applies to this input.

    :param input_id:
            id of the input. This id used to reference this input in other parts of the kervi app.
    :type button_id: str

    :param name:
            Name of this component is used when displayed.
    :type name: str

    :param controller:
            The controller that this component belongs to.
    :type controller: Controller instance.

    :param channel:
            channel on the gpio device that should be used.
    :type channel: int or string.

    :param gpio_device:
            GPIO device to use. Defaults to the host platforms gpio e.g the Raspbery PI's gpio.
    
    :type gpio_device: IGPIODeviceDriver (defined in kervi.utility.hal.gpio).

    """
    def __init__(self, button_id, name, controller, channel, gpio_device=GPIO):
        UINumberControllerInput.__init__(self, button_id, name, controller)
        self.channel = channel
        self.gpio = gpio_device
        self.set_ui_parameter("read_only", True)
        self.gpio.define_as_input(self.channel)
        self.gpio.listen(self.channel, self._on_change)
        self.value = self.gpio.get(self.channel)


    @property
    def value(self):
        """Current state of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        pass

    def _load_persisted(self):
        pass

    def _on_change(self, state):
        self._set_value(state)

class SensorControllerInput(UINumberControllerInput):
    """
    Sensor input that listens to a sensor.
    This input will show up as an UINumberControllerInput when linked to a dashboard panel
    and all ui settings for UINumberControllerInput applies to this input.

    :param input_id:
            id of the input. This id used to reference this input in other parts of the kervi app.
    :type button_id: str

    :param name:
            Name of this component is used when displayed.
    :type name: str

    :param controller:
            The controller that this component belongs to.
    :type controller: Controller instance.

    :param sensor_id:
            id of sensor to listen to.
    :type channel: string.

    """
    def __init__(self, input_id, name, sensor_id, controller):
        UINumberControllerInput.__init__(self, input_id, name, controller)
        self._sensor_id = sensor_id
        self._value = 0
        self.set_ui_parameter("read_only", True)
        self.spine.register_event_handler("NewSensorReading", self._new_sensor_reading, self._sensor_id)
        sensor_value = self.spine.send_query("getSensorValue", self._sensor_id)
        if sensor_value:
            self._set_value(sensor_value)

    def _new_sensor_reading(self,sensor_id, sensor_value):
        self._set_value(sensor_value["value"])

    @property
    def value(self):
        """Current state of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        pass

    def _load_persisted(self):
        pass

class UITextControllerInput(KerviComponent):
    """
    Text input component
    """
    def __init__(self, input_id, name, controller):
        KerviComponent.__init__(self, input_id, "text-input", name)
        #self.spine = Spine()
        self.controller = controller
        self._value = ""
        self.command = self.component_id + ".setValue"
        self.spine.register_command_handler(self.command, self._set_value)
        self.input_type = "text"
        self._ui_parameters = {
            "size": 0,
            "input_size": 50,
            "type": "text",
            "link_to_header": False,
            "label_icon": None,
            "label": self.name,
            "place_holder" : None,
            "flat": False,
            "inline": False
        }
        self._persist_value = False
        self.controller.add_input(self)

    @property
    def input_id(self):
        return self.component_id

    @property
    def value(self):
        """Current state of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        self._set_value(new_value)

    @property
    def persist_value(self):
        """ If true the value is saved to kervi DB when it change"""
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        self._persist_value = do_persist

    def _load_persisted(self):
        if self._persist_value:
            self._set_value(self.settings.retrieve_value("value"))

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.

            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.

            * *input_size* (``int``) -- width of the slider as a percentage of the total container it sits in.


        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            *kwargs
            )

    def _set_value(self, nvalue, key, key_code):
        if self.value != nvalue:
            self.spine.log.debug(
                "value change on input:{0}/{1} value:{2}",
                self.controller.component_id,
                self.component_id,
                nvalue
            )
            old_value = self._value
            self._value = nvalue
            self.value_changed(nvalue, old_value)

            if self.persist_value:
                settings.store_value("value", self.value)

            self.spine.trigger_event(
                "changeControllerInputValue",
                self.component_id,
                {"input":self.component_id, "value":nvalue}
            )

    def value_changed(self, new_value, old_value):
        """
        Abstract method called when the content of the text box change.
        """
        self.controller.input_changed(self)

    def _get_info(self):
        return {
            "value":self.value,
            "command":self.command
        }

    def on_get_value(self):
        return self.value

class UIDateTimeControllerInput(KerviComponent):
    """
    A date and time component.
    """
    def __init__(self, input_id, name, input_type, controller):
        KerviComponent.__init__(self, input_id, "datetime-input", name)
        #self.spine = Spine()
        self.controller = controller
        self.input_type = input_type
        self._value = ""
        self.command = self.component_id + ".setValue"
        self.spine.register_command_handler(self.command, self._set_value)
        self._ui_parameters = {
            "size": 0,
            "input_size": 50,
            "type": "datetime",
            "link_to_header": False,
            "label_icon": None,
            "label": self.name,
            "flat": False,
            "inline": False
        }
        self._persist_value = False
        self.controller.add_input(self)

    @property
    def input_id(self):
        return self.component_id

    @property
    def value(self):
        """Current state of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        self._set_value(new_value)

    @property
    def persist_value(self):
        """ If true the value is saved to kervi DB when it change"""
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        self._persist_value = do_persist

    def _load_persisted(self):
        if self._persist_value:
            self._set_value(self.settings.retrieve_value("value"))

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.

            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.

            * *input_size* (``int``) -- width of the slider as a percentage of the total container it sits in.

            * *type* (``str``) -- Kind of input either 'date' or 'time'.


        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            **kwargs
            )

    def _set_value(self, nvalue):
        if self.value != nvalue:
            self.spine.log.debug(
                "value change on input:{0}/{1} value:{2}",
                self.controller.component_id,
                self.component_id,
                nvalue
            )
            old_value = self.value
            self._value = nvalue
            self.value_changed(nvalue, old_value)
            if self._persist_value:
                settings.store_value("value", self.value)

            self.spine.trigger_event(
                "changeControllerInputValue",
                self.component_id,
                {"input":self.component_id, "value":nvalue}
            )

    def value_changed(self, newValue, oldValue):
        self.controller.input_changed(self)

    def _get_info(self):
        return {
            "inputType": self.input_type,
            "value":self.value,
            "command":self.command
        }

    def on_get_value(self):
        return self.value

class Controller(KerviComponent):
    """
    A Kervi controller is a class that acts upon input from users or events or the underlaying os.

    Examples for controllers are motor control, servo control, output to IO.

    """
    def __init__(self, controller_id, name):
        KerviComponent.__init__(self, controller_id, "controller", name)
        self.components = []
        self._active = True
        self.type = "unknown"
        self.parameters = {}
        self._ui_parameters = {
            "size": 0,
            "type": "normal",
            "link_to_header": False,
            "icon": None,
            "flat": False,
            "inline": False
        }
        self.spine.register_event_handler("processTerminating",self._on_terminate)

    
    @property
    def controller_id(self):
        return self.component_id

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this component to a dashboard section.

        :param dashboard_id:
            id of the dashboard to link to.
        :type section_id: str

        :param section_id:
            id of the section.
        :type section_id: str

        :param \**kwargs:
            Use the kwargs below to override default values set in ui_parameters

        :Keyword Arguments:
            * *ui_size* (``int``) -- Size of the component in dashboard unit size.
                In order to make the sections and components align correct a dashboard unit is defined.
                Default the dashboard unit is a square that is 150 x 150 pixels.
                The width of the select box is ui_size * dashboard unit size.

            * *type* (``str``) -- Type of select box, dropdown or list.
            * *link_to_header* (``str``) -- Add this component to header of section.
            * *icon* (``str``) -- Icon that should be displayed together with label.
            * *flat* (``bool``) -- Flat look and feel.

        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            *kwargs
            )

    def add_input(self, *args):
        for component in args:
            self.components += [component]
            component._load_persisted()

    def _on_terminate(self, id):
        if self._active:
            self._active = False
            try:
                self.spine.log("exit controller:{0}", self.component_id)
            except:
                pass
            finally:
                self.exit()

    def exit(self):
        """
        Abstract method that is called when the kervi application stops.
        """
        pass

    def input_changed(self, changed_input):
        """
        Abstract method that is called by when one of the controller inputs change.

        You can implement this method if your controller logic needs to respond
        changes in multiple inputs.

        :param changed_input:
            Input that has changed. You can read the value of the changed input
            via the inputs value property.

        :type changed_input: Input component
        """
        pass

    def _get_info(self):
        components = []
        for component in self.components:
            components += [component.get_reference()]

        template = None
        try:
            import os.path
            import sys
            modulepath = os.path.dirname(sys.modules[self.__class__.__module__].__file__)
            class_name = self.__class__.__name__
            cpath = os.path.join(modulepath, class_name + ".tmpl.html")
            if os.path.isfile(cpath):
                template_file = open(cpath, 'r')
                template = template_file.read()
        except:
            pass

        return {
            "components":components,
            "template" : template,
            "type": self.type
        }




import time
from datetime import datetime

from kervi.spine import Spine
from kervi.utility.thread import KerviThread
from kervi.utility.component import KerviComponent

VALUE_COUNTER = 0

class DynamicValueList(object):
    def __init__(self, controller, is_input, inject_in_controller=False):
        self._items = {}
        self.is_input = is_input
        self.controller = controller
        self.count = 0
        self._inject = inject_in_controller

    def add(self, value_id, name, value_class):
        item = value_class(
            name,
            input_id=self.controller.component_id + "." + value_id,
            is_input=self.is_input,
            index=self.count
        )

        #if self._inject and self.controller:
        #    setattr(self.controller, value_id, item)
        #setattr(self, value_id, item)
        
        self.count += 1
        self._items[value_id] = item
        if self.is_input:
            item.add_observer(self.controller)
        return item

    @property
    def keys(self):
        return self._items.keys()

    def __getitem__(self, item_id):
        return self._items[item_id]


class DynamicValue(KerviComponent):
    """
    Generic value
    """
    def __init__(self, name, value_type, input_id=None, is_input=True, parent=None, index = None):
        global VALUE_COUNTER
        if input_id is None:
            VALUE_COUNTER += 1
            if parent:
                input_id = parent.component_id + ".DV_" + str(VALUE_COUNTER)
            else:
                input_id = "DV_" + str(VALUE_COUNTER)
        elif parent:
            input_id = parent.component_id + "." + input_id

        KerviComponent.__init__(self, input_id, value_type, name)
        #self.spine = Spine()
        self.unit = ""
        self._index = index
        self.is_input = is_input
        self._value = None
        self._observers = []
        if parent:
            self._observers += [parent]
        self._spine_observers = []
        self.command = self.component_id + ".setValue"
        self.spine.register_command_handler(self.command, self._set_value)
        self.spine.register_query_handler("getDynamicValue", self._query_value)

        self._ui_parameters = {
            "size": 0,
            "input_size": 50,
            "type": "unkonwn",
            "link_to_header": False,
            "label_icon": None,
            "label": self.name,
            "flat": False,
            "inline": False,
            "is_input":self.is_input,
            "value_size":10
        }
        self._persist_value = False

    @property
    def index(self):
        return self._index

    @property
    def input_id(self):
        return self.component_id

    @property
    def value(self):
        """Current value of the component"""
        return self._value

    @value.setter
    def value(self, new_value):
        self._set_value(new_value)

    @property
    def persist_value(self):
        """ If true the value is saved to kervi DB when it change"""
        return self._persist_value

    @persist_value.setter
    def persist_value(self, do_persist):
        self._persist_value = do_persist

    
    def _query_value(self, id):
        if self.component_id == id:
            return self.value
    
    def _load_persisted(self):
        if self._persist_value:
            self._set_value(self.settings.retrieve_value("value"), False)

    def link_to(self, src):
        if isinstance(src, DynamicValue):
            if src.is_input and not self.is_input:
                self.add_observer(src)
            elif not src.is_input and self.is_input:
                src.add_observer(self)
            else:
                raise Exception("input/output mismatch in dynamic value link:{0} - {1}".format(src.name, self.name))
        elif isinstance(src, str):
            if len(self._spine_observers) == 0:
                self.spine.register_event_handler("dynamicValueChanged", self._value_changed_event)
            
            self._spine_observers += [src]

    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.

            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.

            * *input_size* (``int``) -- width of the slider as a percentage of the total container it sits in.

            * *value_size* (``int``) -- width of the value area as a percentage of the total container it sits in.

            * *type* (``string``) -- 'vertical_slider' or 'horizontal_slider.
        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            **kwargs
            )

    def add_observer(self, observer):
        self._observers += [observer]

    def _value_changed_event(self, id, value ):
        if value["id"] in self._spine_observers:
            self.value = value["value"]

    def _set_value(self, nvalue, allow_persist=True):
        if self._value != nvalue:
            self.spine.log.debug(
                "value change on input:{0} value:{1}",
                self.component_id,
                nvalue
            )

            old_value = self.value
            self._value = nvalue
            self.value_changed(nvalue, old_value)
            for observer in self._observers:
                observer.dynamic_value_changed(self)

            if self._persist_value and allow_persist:
                self.settings.store_value("value", self.value)

            timestamp = (datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()
            self.spine.trigger_event(
                "dynamicValueChanged",
                self.component_id,
                {"id":self.component_id, "value":nvalue, "timestamp":timestamp}
            )

    def dynamic_value_changed(self, dynamic_value):
        self.value = dynamic_value.value

    def value_changed(self, new_value, old_value):
        pass

    def _get_info(self):
        return {
            "unit":self.unit,
            "value":self.value,
            "command":self.command,
            "visible": self.visible
        }

    def on_get_value(self):
        return self.value

class DynamicNumber(DynamicValue):
    """
    A number input component shows as a slider on dashboards.
    """
    def __init__(self, name, input_id=None, is_input=True, parent=None, index = None):
        DynamicValue.__init__(self, name, "dynamic-number", input_id, is_input, parent, index)
        #self.spine = Spine()
        self._min_value = -100
        self._max_value = 100
        self.unit = ""
        self._value = 0
        self._delta = None
        self._save_to_db = False
        self._ui_parameters["type"] = "horizontal_slider"
        self._ui_parameters["chart_points"] = 60
        self._ui_parameters["show_sparkline"] = False

        self._upper_fatal_limit = None
        self._upper_warning_limit = None
        self._lower_warning_limit = None
        self._lower_fatal_limit = None

        self._upper_fatal_message = None
        self._upper_warning_message = None
        self._lower_warning_message = None
        self._lower_fatal_message = None

        self._normal_message = None

        self._log_warnings = False
        self._log_errors = False

        self._sparkline = []
        self._last_reading = None
        
    @property
    def persist_to_db(self):
        """
        If true the method new_sensor_reading
        will save the sensor reading to DB.

        :type: ``bool``
        """
        return self._save_to_db

    @persist_to_db.setter
    def persist_to_db(self, value):
        self._save_to_db = value

    @property
    def delta(self):
        """
        Enter how much a sensor value should change to trigger a new sensor read response.
        The method new_sensor_reading triggers db saving.
        Only relevant if persist_to_db is True.
        :type: ``float``
        """
        return self._store_delta

    @delta.setter
    def delta(self, value):
        self._store_delta = value

    @property
    def max(self):
        """
        Maximum value the sensor may mesaure.

        :type: ``float``
        """
        return self._max_value

    @max.setter
    def max(self, value):
        self._max_value = value

    @property
    def min(self):
        """
        Minimum value the sensor may mesaure.

        :type: ``float``
        """
        return self._min_value

    @min.setter
    def min(self, value):
        self._min_value = value

    @property
    def unit(self):
        """
        Mesauring unit for value. Enter values like C, F, hPa

        :type: ``str``
        """
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = value
      

    @property
    def upper_fatal_limit(self):
        """
        If set the sensor will trigger an event *sensorUpperFatal* if the value pases this limit.
        When the sensor is displayed on a dashboard
        the zone from upper_fatal_limit to max is marked red.

        :type: ``float``
        """
        return self._upper_fatal_limit


    @property
    def normal_message(self):
        """
        Message to log when the sensor value reach normal levels from lower or upper limits.

        :type: ``str``
        """
        return self._normal_message

    @normal_message.setter
    def normal_message(self, value):
        self._normal_message = value

    @upper_fatal_limit.setter
    def upper_fatal_limit(self, value):
        self._upper_fatal_limit = value

    @property
    def upper_fatal_message(self):
        """
        Message to log when the sensor value reach upper_fatal_limit

        :type: ``str``
        """
        return self._upper_fatal_message

    @upper_fatal_message.setter
    def upper_fatal_message(self, value):
        self._upper_fatal_message = value

    @property
    def upper_warning_limit(self):
        """
        If set the sensor will trigger an event *sensorUpperWarning* if the value pases this limit.
        When the sensor is displayed on a dashboard the zone from upper_warning_limit
        to upper_fatal_limit or max is marked yellow.

        :type: ``float``
        """
        return self._upper_warning_limit

    @upper_warning_limit.setter
    def upper_warning_limit(self, value):
        self._upper_warning_limit = value

    @property
    def upper_warning_message(self):
        """
        Message to log when the sensor value reach upper_warning_limit

        :type: ``str``
        """
        return self._upper_warning_message

    @upper_warning_message.setter
    def upper_warning_message(self, value):
        self._upper_warning_message = value

    @property
    def lower_warning_limit(self):
        """
        If set the sensor will trigger an event *sensorLowerWarning*
        if the value pases this limit from a higher previus value.
        When the sensor is displayed on a dashboard
        the zone from min or lower_fatal_limit to lower_warning_limit is marked yellow.

        :type: ``float``
        """
        return self._lower_warning_limit

    @lower_warning_limit.setter
    def lower_warning_limit(self, value):
        self._lower_warning_limit = value

    @property
    def lower_warning_message(self):
        """
        Message to log when the sensor value reach lower_warning_limit

        :type: ``str``
        """
        return self._lower_warning_message

    @lower_warning_message.setter
    def lower_warning_message(self, value):
        self._lower_warning_message = value

    @property
    def lower_fatal_limit(self):
        """
        If set the sensor will trigger an event *sensorLowerFatal*
        if the value pases this limit from a higher previus value.
        When the sensor is displayed on a dashboard
        the zone from min to lower_fatal_limit is marked red.

        :type: ``float``
        """
        return self._lower_fatal_limit

    @lower_fatal_limit.setter
    def lower_fatal_limit(self, value):
        self._lower_fatal_limit = value

    @property
    def lower_fatal_message(self):
        """
        Message to log when the sensor value reach lower_fatal_limit

        :type: ``str``
        """
        return self._lower_fatal_message

    @lower_fatal_message.setter
    def lower_fatal_message(self, value):
        self._lower_fatal_message = value

    def _get_info(self):
        return {
            "isInput":self.is_input,
            "unit":self.unit,
            "value":self.value,
            "maxValue":self._max_value,
            "minValue":self._min_value,
            "command":self.command,
            "upperFatalLimit":self.upper_fatal_limit,
            "upperWarningLimit":self.upper_warning_limit,
            "lowerWarningLimit":self.lower_warning_limit,
            "lowerFatalLimit":self.lower_fatal_limit,
            "sparkline":self._sparkline,
        }

    def __delta_exceeded(self, value):
        if self._delta is None:
            return True
        elif value is None:
            return False
        elif self._value is None:
            return True
        elif value >= self._value + self.delta:
            return True
        elif value <= self._value - self.delta:
            return True
        else:
            return False

    def _set_value(self, new_value, allow_persist=True):
        if self.__delta_exceeded(new_value):
            self.spine.log.debug(
                "value change on input:{0} value:{1}",
                self.component_id,
                new_value
            )
            old_value = self._value
            self._value = new_value
            self.value_changed(new_value, old_value)

            for observer in self._observers:
                observer.dynamic_value_changed(self)
            if self._persist_value and allow_persist:
                self.settings.store_value("value", self.value)

            if len(self._sparkline) == 0:
                self._sparkline += [new_value]
            elif len(self._sparkline) >= 10:
                self._sparkline.pop(0)
            self._sparkline += [new_value]
            self._last_reading = time.clock()

            timestamp = (datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()
            val = {"value_id":self.component_id, "value":new_value, "timestamp":timestamp}
            if self.persist_to_db:
                self.spine.send_command("StoreDynamicValue", val)

            self.spine.trigger_event(
                "dynamicValueChanged",
                self.component_id,
                {"id":self.component_id, "value":new_value, "timestamp":timestamp}
            )

    def _send_messages(self, old_val, value):
        if old_val == None:
            return
        if self.lower_fatal_message:
            if old_val > self.lower_fatal_limit and value <= self.lower_fatal_limit:
                self.user_log_message(self.lower_fatal_message, level=1)

        if self.lower_warning_message:
            if self.lower_fatal_limit and self.lower_warning_limit:
                if old_val < self.lower_fatal_limit and value >= self.lower_fatal_limit:
                    self.user_log_message(self.lower_warning_message, level=2)

                elif old_val > self.lower_warning_limit and value <= self.lower_warning_limit:
                    self.user_log_message(self.lower_warning_message, level=2)

            elif self.lower_warning_limit:
                if old_val > self.lower_warning_limit and value <= self.lower_warning_limit:
                    self.user_log_message(self.lower_warning_message, level=2)

        if self.upper_fatal_message:
            if old_val < self.upper_fatal_limit and value >= self.upper_fatal_limit:
                self.user_log_message(self.upper_fatal_message, level=1)

        if self.upper_warning_message:
            if self.upper_fatal_limit and self.upper_warning_limit:
                if old_val > self.upper_fatal_limit and value <= self.upper_fatal_limit:
                    self.user_log_message(self.upper_warning_message, level=2)
                elif old_val < self.upper_warning_limit and value >= self.upper_warning_limit:
                    self.user_log_message(self.upper_warning_message, level=2)


            elif self.upper_warning_limit:
                if old_val < self.upper_warning_limit and value >= self.upper_warning_limit:
                    self.user_log_message(self.upper_warning_message, level=2)

        if self.normal_message:
            if self.upper_warning_limit:
                if old_val >= self.upper_warning_limit and value < self.upper_warning_limit and value:
                    self.user_log_message(self.normal_message, level=3)
            elif self.upper_fatal_limit:
                if old_val >= self.upper_fatal_limit and value < self.upper_fatal_limit and value:
                    self.user_log_message(self.normal_message, level=3)

            if self.lower_warning_limit:
                if old_val <= self.lower_warning_limit and value > self.lower_warning_limit and value:
                    self.user_log_message(self.normal_message, level=3)
            elif self.lower_fatal_limit:
                if old_val <= self.lower_fatal_limit and value > self.lower_fatal_limit and value:
                    self.user_log_message(self.normal_message, level=3)

    
    def link_to_dashboard(self, dashboard_id, section_id, **kwargs):
        r"""
        Links this input to a dashboard panel.

        :param dashboard_id:
            Id of the dashboard to link to.
        :type dashboard_id: str

        :param panel_id:
            Id of the panel on the dashboard to link to.
        :type panel_id: str

        :param \**kwargs:
            Use the kwargs below to override default values for ui parameters

        :Keyword Arguments:
            * *size* (``int``) -- The number of dashboard cells the input should occupy horizontal.
                If size is 0 (default) the input and label will expand to the width of the panel.

            * *link_to_header* (``str``) -- Link this input to header of the panel.

            * *label_icon* (``str``) -- Icon that should be displayed together with label.

            * *label* (``str``) -- Label text, default value is the name of the input.

            * *flat* (``bool``) -- Flat look and feel.

            * *inline* (``bool``) -- Display input and label in its actual size
                If you set inline to true the size parameter is ignored.
                The input will only occupy as much space as the label and input takes.

            * *input_size* (``int``) -- width of the slider as a percentage of the total container it sits in.

            * *value_size* (``int``) -- width of the value area as a percentage of the total container it sits in.

            * *type* (``string``) -- 'vertical_slider' or 'horizontal_slider.
            
        """
        KerviComponent.link_to_dashboard(
            self,
            dashboard_id,
            section_id,
            **kwargs
            )


class DynamicString(DynamicValue):
    """
    Text input component
    """
    def __init__(self, name, input_id=None, is_input=True, parent=None, index = None):
        DynamicValue.__init__(self, name, "dynamic-string", input_id, is_input, parent, index)
        #self.spine = Spine()
        self._value = ""
        self._ui_parameters["type"] = "text"

class DynamicDateTime(DynamicValue):
    """
    A date and time component.
    """
    def __init__(self, name, input_type="datetime", input_id=None, index = None):
        DynamicValue.__init__(self, name, "dynamic-datetime", input_id, index)
        #self.spine = Spine()
        self._value = ""
        self._ui_parameters["type"] = input_type

class DynamicBoolean(DynamicValue):
    """
    Switch button controller component, shows a on/off button in UI
    """
    def __init__(self, name, input_id=None, is_input=True, parent=None, index = None):
        DynamicValue.__init__(self, name, "dynamic-boolean", input_id, is_input, parent, index)
        self._value = False
        self._ui_parameters["type"] = "switch"
        self._ui_parameters["on_text"] = "On"
        self._ui_parameters["off_text"] = "Off"
        self._ui_parameters["on_icon"] = None
        self._ui_parameters["off_icon"] = None
        self._ui_parameters["button_icon"] = None
        self._ui_parameters["button_text"] = self.name


class DynamicEnum(DynamicValue):
    r"""
    EnumControllerValue holds a selection of predefined values to select amount.

    Usage:

    .. code-block:: python

        class CameraFrameRate(EnumControllerValue):
            def __init__(self, controller):
            EnumControllerValue.__init__(self, controller.controllerId+".framerate", "Framerate" , controller)
            self.addOption("5", "5 / sec")
            self.addOption("10", "10 / sec")
            self.addOption("15", "15 / sec", True)

        def change(self, selectedOptions):
            print("Frame rate changed", selectedOptions)

    """
    def __init__(self, name, input_id=None, is_input=True, parent=None, index = None):
        DynamicValue.__init__(self, name, "dynamic-enum", input_id, is_input, parent, index)

        self.options = []
        self.selected_options = []
        self._ui_parameters["type"] = "dropdown"

    def add_option(self, value, text, selected=False):
        """
        Add option to select
        """
        option = {"value": value, "text": text, "selected":selected}
        self.options += [option]
        if selected:
            self.selected_options += [option]
            if self._persist_value:
                self.settings.store_value("selected_options", self.selected_options)

    def _get_info(self):
        return {
            "options": self.options,
            "command":self.command
        }

    def _set_value(self, selected_options, allow_persist=True):
        self.spine.log.debug(
            "enum value changed:{0}",
            self.component_id
        )

        for option in self.options:
            option["selected"] = False

        self.selected_options = []
        for selected_option in selected_options:
            for option in self.options:
                if option["value"] == selected_option:
                    option["selected"] = True
                    self.selected_options += [option]

        self.value_changed(self.selected_options, None)

        if self._persist_value and allow_persist:
            self.settings.store_value("value", self.selected_options)

        self.spine.trigger_event(
            "dynamicValueChanged",
            self.component_id,
            {"select":self.component_id, "value":self.options}
        )



# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT

""" CQRS handles in process communication """
#import sys
#import traceback
#import time
import threading
import inspect
from kervi.utility.named_lists import NamedLists
from kervi.utility.thread import KerviThread

try:
    import Queue
except ImportError:
    import queue as Queue
from  kervi.kervi_logging import KerviLog

class _QueryThread(threading.Thread):
    def __init__(self, handler, query, args, spine, **kwargs):
        threading.Thread.__init__(self)
        injected = kwargs.get("injected", "")
        scope = kwargs.get("scope", "global")
        
        self.daemon = True
        self.handler = handler
        self.query = query
        self.args = args
        self.result = []
        self.injected = injected
        self.scope = scope

        
        import traceback
        #spine.log.debug("qtx:{0}, handler:{1}", self.query, self.handler)
        #for line in traceback.format_stack():
        #    spine.log.debug(line.strip())
        
    def run(self):
        self.result = self.handler(self.query, self.args, injected=self.injected, scope=self.scope)

class _CQRSQueue(KerviThread):
    def __init__(self, name):
        KerviThread.__init__(self)
        self.queues = [Queue.Queue(), Queue.Queue(), Queue.Queue()]
        self.name = name

    def get_info(self):
        queueSize = []
        for queue in self.queues:
            queueSize += [len(queue)]

        info = {"queueSize":queueSize}
        return info

    def set_queue_handler(self, handler):
        self.queue_handler = handler
        super(KerviThread, self).start()

    def add(self, item, priorty=0):
        self.queues[0].put_nowait(item)

    def _step(self):
        item = self.queues[0].get()
        self.queue_handler(item)

class _CQRSBus(object):
    applicationId = ""
    cmd_handlers = NamedLists()
    event_handlers = NamedLists()
    query_handlers = NamedLists()

    cmd_queue = _CQRSQueue("cmd")
    query_queue = _CQRSQueue("query")
    event_queue = None #CQRSQueue("event")

    linked_spines = []
    log = None


    def __init__(self):
        return

    def stop(self):
        self.cmd_queue.stop()
        self.query_queue.stop()
        self.event_queue.stop()

    def reset(self):
        self.cmd_handlers = NamedLists()
        self.event_handlers = NamedLists()
        self.query_handlers = NamedLists()
        self.cmd_queue.stop()
        self.query_queue.stop()
        #self.eventQueue.stop()

        self.cmd_queue = _CQRSQueue("cmd")
        self.query_queue = _CQRSQueue("query")
        self.event_queue = _CQRSQueue("event")

        self.linked_spines = []

    def set_log(self, logName):
        self.log = KerviLog(logName)

    def add_linked_spine(self, linkedspine):
        self.linked_spines += [linkedspine]

    def get_queue_info(self):
        info = {
            "cmd":self.cmd_queue.get_info(),
            "queryQueue":self.query_queue.get_info(),
            "event":self.event_queue.get_info()
        }
        return info

    def start_queues(self):
        self.cmd_queue.set_queue_handler(self.command_queue_handler)
        self.event_queue.set_queue_handler(self.event_queue_handler)

    def register_command_handler(self, name, func, **kwargs):
        injected = kwargs.get("injected", "")
        self.log.debug("register command handler, command:{0} injected:{1}", name, injected)
        self.cmd_handlers.add(name, func)
        for linked_spine in self.linked_spines:
            linked_spine.add_linked_command_handler(name, injected=injected)

    def register_event_handler(self, name, func, id=None, **kwargs):
        injected = kwargs.get("injected", "")
        self.log.debug("register event handler event:{0} id:{1} injected:{2}", name, id, injected)
        if id:
            self.event_handlers.add(name+"/"+id, func)
        else:
            self.event_handlers.add(name, func)

        for linkedspine in self.linked_spines:
            linkedspine.add_linked_event_handler(name, id, injected=injected)

        return func

    def register_query_handler(self, name, func, **kwargs):
        injected = kwargs.get("injected", "")
        self.log.debug("register query handler query:{0} injected:{1}", name, injected)
        self.query_handlers.add(name, func)
        for linked_spine in self.linked_spines:
            linked_spine.add_linked_query_handler(name, injected=injected)

    def send_command(self, command, *args, **kwargs):
        injected = kwargs.get("injected", "")
        scope = kwargs.get("scope", "global")
        self.log.debug("sendcommand:{0} injected:{1} scope:{2}", command, injected, scope)
        self.cmd_queue.add(
            {"command":command, "args":args, "injected":injected, "scope":scope}, kwargs.get("priority", 2)
        )

    def command_queue_handler(self, queue_item):
        func_list = self.cmd_handlers.get_list_data(queue_item['command'])
        if func_list:
            for func_handler in func_list:
                try:
                    argspec = inspect.getargspec(func_handler)
                    if not argspec.keywords:
                        func_handler(*queue_item['args'])
                    else:
                        func_handler(*queue_item['args'], injected=queue_item["injected"], scope=queue_item["scope"])
                except:
                    self.log.exception("commandQueueHandler error:"+queue_item['command'])

    def send_query(self, query, *args, **kwargs):
        injected = kwargs.get("injected", "")
        scope = kwargs.get("scope", "global")
        self.log.debug("sendQuery:{0} injected:{1} scope:{2}", query, injected, scope)
        if query == "getQueueInfo":
            return self.get_queue_info()
        query_thread = _QueryThread(self.query_handler, query, args, self, injected=injected, scope=scope)
        self.log.debug("sendQuery thread start:{0}", query_thread)

        query_thread.start()
        query_thread.join(3)
        if not query_thread.isAlive():
            self.log.debug("sendQuery thread done:{0}", query_thread)
            return query_thread.result
        else:
            self.log.debug("sendQuery thread timeout, query:{0}", query)
            return []

    def query_handler(self, query, args, **kwargs):
        injected = kwargs.get("injected", "")
        scope = kwargs.get("scope", "global")

        #self.log.debug("query handler called:{0} injected:{1}", query, injected)
        func_list = self.query_handlers.get_list_data(query)
        self.log.debug("query handler called:{0} injected:{1}, func:{2}", query, injected, func_list)
        result = []
        try:
            if func_list:
                for func in func_list:
                    argspec = inspect.getargspec(func)
                    if not argspec.keywords:
                        sub_result = func(*args)
                        if sub_result:
                            result += [sub_result]
                    else:
                        sub_result = func(*args, injected=injected, scope=scope)
                        if sub_result:
                            result += [sub_result]
            if len(result) == 1:
                return result[0]
        except:
            self.log.exception("Exception in query:" + query)

        return result

    def trigger_event(self, event, id, *args, **kwargs):
        injected = kwargs.get("injected", "")
        scope = kwargs.get("scope", "global")
        self.log.debug("triggerEvent:{0}, id:{1} injected:{2}, scope:{3}", event, id, injected, scope)
        self.event_queue.add(
            {'event':event, 'id':id, 'args':args, "injected":injected, "scope":scope},
            kwargs.get("priority", 2)
        )

    def event_queue_handler(self, queue_item):
        func_list = self.event_handlers.get_list_data(queue_item['event'])
        try:
            if func_list:
                for func in func_list:
                    argspec = inspect.getargspec(func)
                    if not argspec.keywords:
                        func(None, *queue_item['args'])
                    else:
                        func(None, *queue_item['args'], injected=queue_item["injected"], scope=queue_item["scope"])
            if queue_item["id"]:
                func_list = self.event_handlers.get_list_data(queue_item['event']+"/"+queue_item['id'])
                if func_list:
                    for func in func_list:
                        argspec = inspect.getargspec(func)
                        if not argspec.keywords:
                            func(queue_item["id"], *queue_item['args'])
                        else:
                            func(queue_item["id"], *queue_item['args'], injected=queue_item["injected"], scope=queue_item["scope"])
        except:
            self.log.exception("Exception in event:{0}",queue_item)
    
    def get_commands(self):
        return self.cmd_handlers.get_list_names()

    def get_queries(self):
        return self.query_handlers.get_list_names()

    def get_events(self):
        return self.event_handlers.get_list_names()

    def get_query_handlers(self, query):
        return self.query_handlers.get_list_data(query)

    def get_command_handlers(self, command):
        return self.cmd_handlers.get_list_data(command)

    def get_event_handlers(self, event, id=None):
        func_list = self.event_handlers.get_list_data(event)
        if id:
            func_list += self.event_handlers.get_list_data(event+"/"+id)
        return func_list




# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT
"""Hardware abstraction layer"""
import time
from kervi.hal import gpio
from kervi.hal import i2c
from kervi.spine import Spine
from kervi.utility.thread import KerviThread
import pip
import importlib

_DRIVER = None
GPIO = None

def _load():
    global GPIO, _DRIVER

    if not _DRIVER:
        installed_packages = pip.get_installed_distributions()
        flat_installed_packages = [package.project_name for package in installed_packages]
        known_drivers = [
            ("kervi-hal-win", "kervi_hal_win"),
            ("kervi-hal-linux", "kervi_hal_linux"),
            ("kervi-hal-rpi", "kervi_hal_rpi"),
            ("kervi-hal-generic", "kervi_hal_generic")
        ]
        for driver_name, module_name in known_drivers:
            if driver_name in flat_installed_packages:
                _DRIVER = importlib.import_module(module_name)
                GPIO = get_gpio()
                return driver_name


def get_gpio(gpio_type=None):
    if gpio_type == None:
        return _DRIVER.get_gpio_driver()



def default_i2c_bus():
    if _DRIVER:
        return _DRIVER.default_i2c_bus()
    return 0

def i2c(address, bus=default_i2c_bus()):
    return _DRIVER.get_i2c_driver(address, bus)



def get_camera_driver(source = None):
    return _DRIVER.get_camera_driver(source)

#if not _DRIVER:
#    _load()


class SensorDeviceDriver(object):

    @property
    def dimensions(self):
        return 1

    @property
    def dimension_labels(self):
        return []

    @property
    def device_name(self):
        return None

    @property
    def max(self):
        return None

    @property
    def min(self):
        return None

    @property
    def type(self):
        raise NotImplementedError

    @property
    def unit(self):
        raise NotImplementedError

    def read_value(self):
        raise NotImplementedError

    @property
    def logger(self):
        return Spine().log

class I2CaddressOutOfBoundsError(Exception):
    def __init__(self, device_name, address):
        super(I2CaddressOutOfBoundsError, self).__init__(
            '{0} I2C Address Out of Bounds:{1}'.format(device_name, address)
        )

class DeviceChannelOutOfBoundsError(Exception):
    def __init__(self, device_name, channel):
        super(DeviceChannelOutOfBoundsError, self).__init__(
            '{0} Exception: Channel Out of Bounds, channel={1}'.format(device_name, channel)
        )

# Exception class for a DAC value out of bounds
class DACValueOutOfBoundsError(Exception):
    def __init__(self, device_name, channel, value):
        super(DACValueOutOfBoundsError, self).__init__(
            '{0} Exception: DAC Output Value Out of Bounds, channel={1} value={2}'.format(device_name, channel, value)
        )

class I2CSensorDeviceDriver(SensorDeviceDriver):
    def __init__(self, address, bus):
        if address > 0x77:
            raise I2CaddressOutOfBoundsError(self, device_name, address)
        self.i2c = i2c(address, bus)

class I2CGPIODeviceDriver(gpio.IGPIODeviceDriver):
    def __init__(self, address, bus, gpio_id):
        gpio.IGPIODeviceDriver.__init__(self, gpio_id)
        if address > 0x77:
            raise I2CaddressOutOfBoundsError(self, device_name, address)
        self.i2c = i2c(address, bus)

    @property
    def device_name(self):
        return None

    @property
    def num_gpio(self):
        return 0

    def _bit2(self, src, bit, val):
        bit = 1 << bit
        return (src | bit) if val else (src & ~bit)


    def _validate_channel(self, channel):
        # Raise an exception if pin is outside the range of allowed values.
        if channel < 0 or channel >= self.num_gpio:
            raise DeviceChannelOutOfBoundsError(self.device_name, channel)

class ChannelPollingThread(KerviThread):
    def __init__(self, channel, device, callback, polling_time=.1):
        KerviThread.__init__(self)
        self._callback = callback
        self._channel = channel
        self._device = device
        self._value = None
        self._polling_time = polling_time
        self.alive = False
        self.spine = Spine()
        if self.spine:
            self.spine.register_command_handler("startThreads", self._start_command)
            self.spine.register_command_handler("stopThreads", self._stop_command)

    def _step(self):
        """Private method do not call it directly or override it."""
        try:
            new_value = self._device.get(self._channel)
            if new_value != self._value:
                self._callback(new_value)
                self._value = new_value
            time.sleep(self._polling_time)
        except:
            self.spine.log.exception("_PollingThread")

    def _start_command(self):
        if not self.alive:
            self.alive = True
            KerviThread.start(self)

    def _stop_command(self):
        if self.alive:
            self.alive = False
            self.stop()




# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT

""" general DB handling in Kervi, store sensor values  """
import os
import time
from datetime import datetime
import json
import inspect
import sqlite3 as lite
import kervi.spine as spine
from kervi.utility.thread import KerviThread
0
SPINE = spine.Spine()
FILE_CON = None
MEMORY_CON = None



class MemoryCleanThread(KerviThread):
    def __init__(self):
        KerviThread.__init__(self)
        self.spine = spine.Spine()
        self.alive = False
        if self.spine:
            self.spine.register_command_handler("startThreads", self._start_command)
            self.spine.register_command_handler("stopThreads", self._stop_command)

    def _step(self):
        time.sleep(20)
        try:
            cursor = MEMORY_CON.cursor()
            cursor.execute("DELETE FROM sensorData WHERE id IN (SELECT id FROM sensorData ORDER BY id ASC LIMIT 1000);")
        except lite.Error as msg:
            SPINE.log.error("clean memory db, Command skipped: {0}, command{1}", msg)

    def _start_command(self):
        print("start mcl")
        if not self.alive:
            self.alive = True
            KerviThread.start(self)

    def _stop_command(self):
        if self.alive:
            self.alive = False
            self.stop()

    



DB_CREATE_SQL = """
        CREATE TABLE `log` (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `sourceId`	TEXT,
            `sourceName`	TEXT,
            `area`	TEXT,
            `data`	TEXT,
            `level`	INTEGER,
            `topic`	TEXT,
            `body`	TEXT,
            'ts'    `timeStamp`	REAL,
            `logType`	TEXT
        );
        CREATE TABLE "sensorData" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `sensor`	TEXT,
            `value`	TEXT,
            `timeStamp`	REAL
        );
        CREATE INDEX `sensorindex` ON `sensorData` (`sensor` ,`timeStamp` );
        CREATE TABLE "settings" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `setting_group`	TEXT,
            `name`	TEXT,
            `value`	TEXT
        );
        CREATE INDEX `settingsindex` ON `settings` (`setting_group` ,`name` );
        CREATE TABLE "cron" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `job_id` TEXT,
            `component_id` TEXT
            `name`	TEXT,
            `job_parameters` TEXT
        );
        CREATE TABLE "cron_meta" (
            `id`	INTEGER PRIMARY KEY AUTOINCREMENT,
            `cron_id`	INTEGER,
            `repeat_start` REAL NOT NULL,
            `repeat_end` REAL NOT NULL,
            `repeat_interval` varchar(255) NOT NULL,
            `repeat_year` varchar(255) NOT NULL,
            `repeat_month` varchar(255) NOT NULL,
            `repeat_day` varchar(255) NOT NULL,
            `repeat_week` varchar(255) NOT NULL,
            `repeat_weekday` varchar(255) NOT NULL
        );
        """


class _ObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "to_json"):
            return self.default(obj.to_json())
        elif hasattr(obj, "__dict__"):
            data = dict(
                (key, value)
                for key, value in inspect.getmembers(obj)
                if not key.startswith("__")
                and not inspect.isabstract(value)
                and not inspect.isbuiltin(value)
                and not inspect.isfunction(value)
                and not inspect.isgenerator(value)
                and not inspect.isgeneratorfunction(value)
                and not inspect.ismethod(value)
                and not inspect.ismethoddescriptor(value)
                and not inspect.isroutine(value)
            )
            return self.default(data)
        return obj


def execute_sql(con, sql):
    sql_commands = sql.split(';')

    for command in sql_commands:
        try:
            con.execute(command)
        except lite.Error as msg:
            SPINE.log.error("create db, Command skipped: {0}, command{1}", msg, command)

def init_db():
    global FILE_CON, MEMORY_CON

    MEMORY_CON = lite.connect("kervi_mem.db", check_same_thread=False)
    FILE_CON = lite.connect('kervi.db', check_same_thread=False)

    for con in [FILE_CON, MEMORY_CON]:
        cursor = con.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables_exists = False
        tables = cursor.fetchall()
        for table in tables:
            tables_exists = True
            break

        if not tables_exists:
            execute_sql(con, DB_CREATE_SQL)        

    MemoryCleanThread()

init_db()
TS_START = datetime.utcnow()

def store_sensor_reading(sensor_id, sensor_value, persist=False):
    global FILE_CON
    try:
        if not persist:
            cur = MEMORY_CON.cursor()
        else:
            cur = FILE_CON.cursor()

        cur.execute(
            "INSERT INTO sensorData ('sensor','value','timeStamp')  VALUES (?, ?, ?)",
            (sensor_value["sensor"], sensor_value["value"], sensor_value["timestamp"])
        )
        FILE_CON.commit()
    except lite.Error as er:
        SPINE.log.error('error store sensordata:{0}', er)


def store_setting(group, name, value):
    try:
        setting = retrieve_setting_db(group, name)
        json_value = json.dumps(value, cls=_ObjectEncoder, ensure_ascii=False).encode('utf8')
        if setting:
            cur = FILE_CON.cursor()
            cur.execute(
                "update settings set value=? where id=?",
                (json_value, setting["id"])
            )
            FILE_CON.commit()
        else:
            con = lite.connect('kervi.db')
            cur = con.cursor()
            cur.execute(
                "INSERT INTO settings ('setting_group','name','value')  VALUES (?, ?, ?)",
                (group, name, json_value)
            )
            con.commit()

    except lite.Error as er:
        SPINE.log.error('error store settings data:{0}', er)

def retrieve_setting_db(group, name):
    cur = FILE_CON.cursor()
    cur.execute(
        "select * from settings where setting_group=? and name=?",
        (group, name)
    )

    all_rows = cur.fetchall()
    if len(all_rows) > 0:
        return {
            "id": all_rows[0][0],
            "group": all_rows[0][1],
            "name": all_rows[0][2],
            "value": json.loads(all_rows[0][3])
        }
    return None

def retrieve_setting(group, name):
    setting = retrieve_setting_db(group, name)
    if setting:
        return setting["value"]

def get_sensor_data(sensor, date_from=None, date_to=None, limit=60):
    SPINE.log.debug("getSensorData sensor:{0}, from {1} to {2}", sensor, date_from, date_to)
    if date_from is None:
        date_from = TS_START

    if date_to is None:
        date_to = datetime.utcnow()

    date_from_ts = (date_from - datetime(1970, 1, 1)).total_seconds()
    date_to_ts = (date_to - datetime(1970, 1, 1)).total_seconds()

    try:
        con_list = [MEMORY_CON, FILE_CON]
        result = []
        for con in con_list:
            cur = con.cursor()
            cur.execute(
                "select * from sensorData where sensor=? and timestamp > ? and timestamp < ? LIMIT ?",
                (sensor, date_from_ts, date_to_ts, limit)
            )

            all_rows = cur.fetchall()
            for row in all_rows:
                result += [
                    {
                        "value":row[2],
                        "ts": datetime.fromtimestamp(row[3]).strftime('%Y-%m-%dT%H:%M:%SZ')
                    }
                ]
            if len(result):
                break
        return result
    except lite.Error as er:
        SPINE.log.error('error get sensordata:{0}', er)

def store_log_item(id, item):
    try:
        if not item["persist"]:
            con = MEMORY_CON
        else:
            con = FILE_CON

        cur = con.cursor()
        cur.execute(
            "INSERT INTO log ('sourceId','sourceName','area', 'data', 'level', 'topic', 'body', 'ts', 'logType')  VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
            (item["source_id"], item["source_name"], item["area"], item["data"], item["level"], item["topic"], item["body"], item["timestamp"], item["type"])
        )
        con.commit()
    except lite.Error as er:
        SPINE.log.error('error store log data:{0}', er)

def get_log_sort_field(item):
    return item["timestamp"]

def get_log_items(page, page_size, filters=None):
    connections = [FILE_CON, MEMORY_CON]
    result = []
    for con in connections:
        cur = con.cursor()
        cur.execute(
            "select * from log order by ts desc limit ?, ?",
            (page_size*page, page_size)
        )
        all_rows = cur.fetchall()
        for row in all_rows:
            if row[8]:
                ts = datetime.fromtimestamp(row[8]).strftime('%Y-%m-%dT%H:%M:%SZ')
            else:
                ts = None
            result += [
                {
                    "source_id": row[1],
                    "source_name": row[2],
                    "area": row[3],
                    "data": row[4],
                    "level": row[5],
                    "topic": row[6],
                    "body": row[7],
                    "timestamp": row[8],
                    "type": row[9]
                }
            ]
    #result = sorted(result, get_log_sort_field)
    return result


def create_cron_job(component_id, job_id, name, job_parameters, start, end, repeat_interval, year,month):
    pass

def delete_cron_job():
    pass

def query_cron_job():
    pass

if (SPINE):
    SPINE.register_query_handler("getSensorData", get_sensor_data)
    SPINE.register_command_handler("storeSetting", store_setting)
    SPINE.register_query_handler("retrieveSetting", retrieve_setting)
    SPINE.register_query_handler("getLogItems", get_log_items)
    SPINE.register_event_handler("NewSensorReading", store_sensor_reading)
    SPINE.register_event_handler("userLogMessage", store_log_item)
    SPINE.register_command_handler("createCronJob", create_cron_job)
    SPINE.register_command_handler("deleteCronJob", delete_cron_job)
    SPINE.register_query_handler("queryCronJob", query_cron_job)



# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT
from kervi.spine import Spine

class Settings(object):
    """
    Class that persists settings to the Kervi database.

    :param group:
            To avoid name clash with other settings in the Kervi application
             enter name to group your settings under.

    :type group: ``str``

    """
    def __init__(self, settings_group=None):
        self.group = settings_group
        self.spine = Spine()

    def store_value(self, name, value):
        """Store a value to DB"""
        self.spine.send_command("storeSetting", self.group, name, value)


    def retrieve_value(self, name):
        """Retrieve a value from DB"""
        return self.spine.send_query("retrieveSetting", self.group, name)



# Copyright (c) 2016, Tim Wentzlau
# Licensed under MIT


""" 
This module bootstraps the cental messagegin system in Kervi.
Include this module in all modules where communication is needed.

Usage:
from kervi.spine import Spine

myValue=10
spine = Spine()
spine.sendCommand("SetMyValue",myValue)
"""

from kervi.utility.bus import _CQRSBus

class _KerviSpine(_CQRSBus):
    def version(self):
        return 1.0

S = None
def _init_spine(spine_name="roboSys"):
    global S
    S = _KerviSpine()
    S.set_log(spine_name)
    S.reset()
    S.start_queues()

def Spine():
    """
    Return instance of core communication class in Kervi, ready to use.
    Use this when communication is needed between Kervi sensors, controller
    and other code parts Include this module in all modules where communication is needed.

    Usage:

    myValue=10
    spine = Spine()
    spine.sendCommand("SetMyValue",myValue)
    """
    return S



""" Setup module for kervi generating setup package used with pip """
from distutils.core import setup
import distutils
from kervi.version import VERSION

try:
    distutils.dir_util.remove_tree("dist")
except:
    pass

setup(
    name='kervi',
    packages=[
        'kervi',
        'kervi/utility',
        'kervi/hal'
    ],
    version=VERSION,
    description="""
    A python framework for creating robotic and automation applications on Raspbery pi (and other platforms).
    UI is web based and generated on the fly based on configuration in python code.
    """,
    author='Tim Wentzlau',
    author_email='tim.wentzlau@gmail.com',
    url='https://github.com/kervi/kervi',
    download_url='https://github.com/kervi/kervi/archive/v1.0-alpha.tar.gz',
    keywords=['raspberry pi', 'robotic', 'automation'],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Software Development :: Documentation",
        "Topic :: System :: Monitoring",
        "Environment :: Console",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.4"
    ],
    install_requires=[
        'kervi-ui',
        'kervi-cli',
        'kervi-hal-generic',
        'kervi-device-library',
        'autobahn',
        'pillow'
    ],
    extras_require={
        'rpi':  ['kervi-hal-rpi'],
        'raspberry':  ['kervi-hal-rpi'],
        'win':  ['kervi-hal-win'],
        'windows':  ['kervi-hal-win'],
        'linux':  ['kervi-hal-linux'],
        'mac':  ['kervi-hal-linux'],
    },
    include_package_data=True,
    package_data={
        "kervi":[
            "fonts/*.*"
        ]
    }
)

